/* Copyright 2009-2011 Yorba Foundation
 *
 * This software is licensed under the GNU Lesser General Public License
 * (version 2.1 or later).  See the COPYING file in this distribution.
 */

namespace Publishing.Accounts {

public class SharingAccount {
    private Ag.AccountService account_service = null;

    public SharingAccount(Ag.AccountService account_service) {
        this.account_service = account_service;
    }

    public Signon.AuthSession create_auth_session() throws GLib.Error {
        var auth_data = account_service.get_auth_data();

        debug("Signon-id: %u", auth_data.get_credentials_id());

        return new Signon.AuthSession(auth_data.get_credentials_id(),
                                      auth_data.get_method());
    }

    public HashTable<string,Value?> get_session_parameters(out string mechanism) {
        var auth_data = account_service.get_auth_data();
        mechanism = auth_data.get_mechanism();
        return auth_data.get_parameters();
    }
}

public class SharingAccounts {
    private static Ag.Manager manager = null;
    private string provider_name;
    private Ag.AccountService[] all_accounts;

    public SharingAccounts(string provider_name) {
        if (manager == null) {
            manager = new Ag.Manager.for_service_type("shotwell-sharing");
        }
        manager.enabled_event.connect(on_account_enabled);

        this.provider_name = provider_name;
        all_accounts = get_accounts();
    }

    private void on_account_enabled(uint account_id) {
        /* To keep the implementation simple, just rebuild the account
         * list from scratch */
        all_accounts = get_accounts();
    }

    private Ag.AccountService[] get_accounts() {
        GLib.List<Ag.AccountService> accounts =
            manager.get_enabled_account_services();

        Ag.AccountService[] list = {};

        foreach (Ag.AccountService account_service in accounts) {
            Ag.Account account = account_service.get_account();
            if (account.get_provider_name() == provider_name) {
                list += account_service;
            }
        }
        return list;
    }

    public bool has_enabled_accounts() {
        return all_accounts.length > 0;
    }

    public string[] list_account_names() {
        string[] names = {};
        foreach (Ag.AccountService account_service in all_accounts) {
            names += account_service.get_account().get_display_name();
        }
        return names;
    }

    public SharingAccount? find_account(string? account_name) {
        foreach (Ag.AccountService account_service in all_accounts)
            if (account_service.get_account().get_display_name() == account_name) {
                return new SharingAccount(account_service);
            }

        return null;
    }
}

public abstract class UOAPublishingService : Object, Spit.Pluggable, Spit.Publishing.Service {
    private SharingAccounts account_manager;

    public UOAPublishingService(string provider_name) {
        account_manager = new SharingAccounts(provider_name);
    }

    public virtual int get_pluggable_interface(int min_host_interface, int max_host_interface) {
        return Spit.negotiate_interfaces(min_host_interface, max_host_interface,
            Spit.Publishing.CURRENT_INTERFACE);
    }

    public abstract unowned string get_id();

    public abstract unowned string get_pluggable_name();

    public abstract void get_info(ref Spit.PluggableInfo info);

    public virtual void activation(bool enabled) {
    }

    public abstract Spit.Publishing.Publisher create_publisher(string? account_name, Spit.Publishing.PluginHost host);

    public abstract Spit.Publishing.Publisher.MediaType get_supported_media();

    public SharingAccount? find_account(string? account_name) {
        return account_manager.find_account(account_name);
    }

    public bool is_enabled() {
        return account_manager.has_enabled_accounts();
    }

    public string[] list_account_names() {
        return account_manager.list_account_names();
    }
}

public class UOAPublisherAuthenticator : Object {
    private weak Spit.Publishing.PluginHost host = null;
    private Signon.AuthSession auth_session = null;
    private SharingAccount account = null;
    private bool firstLoginAttempt = true;
    private string welcome_message = null;

    public UOAPublisherAuthenticator(SharingAccount account,
                                     Spit.Publishing.PluginHost host,
                                     string welcome_message)
    {
        this.host = host;
        this.account = account;
        this.welcome_message = welcome_message;
    }

    public signal void authenticated(owned HashTable<string,Value?> session_data);

    public void authenticate() {
        debug("ACTION: authentication requested.");

        do_authentication();
    }

    public HashTable<string,Value?>? get_authentication_data() {
        if (account == null) return null;
        string mechanism = null;
        return account.get_session_parameters(out mechanism);
    }

    private void do_authentication() {
        debug("ACTION: authenticating.");

        HashTable<string,Value?> data = null;
        string mechanism = null;

        if (account != null) {
            try {
                auth_session = account.create_auth_session();
                data = account.get_session_parameters(out mechanism);
                debug("Got account data");
            } catch (GLib.Error e) {
                warning("EVENT: couldn't create session for account: %s",
                        e.message);
            }
        }

        if (data == null) {
            warning ("No account authentication data");
            host.post_error(new Spit.Publishing.PublishingError.SERVICE_ERROR(
                "Error while accessing the account"));
            return;
        }

        if (firstLoginAttempt) {
            firstLoginAttempt = false;
            data.insert("UiPolicy", (int)Signon.SessionDataUiPolicy.NO_USER_INTERACTION);
        } else {
            var windowId = host.get_dialog_xid();
            if (windowId != 0) {
                data.insert("WindowId", (uint)windowId);
            }
        }

        auth_session.process(data, mechanism, on_processed);
        host.set_service_locked(true);
    }

    private void on_processed(Signon.AuthSession self, owned HashTable<string,Value?>? session_data, GLib.Error error){
        host.set_service_locked(false);
        if (error != null) {
            debug("got error: %s", error.message);
            if (error is Signon.Error.USER_INTERACTION) {
                debug("User interaction!");
                do_show_service_welcome_pane();
            } else {
                host.post_error(new Spit.Publishing.PublishingError.SERVICE_ERROR("Authentication failed"));
            }
            return;
        }

        authenticated(session_data);
    }

    private void do_show_service_welcome_pane() {
        debug("ACTION: showing service welcome pane.");

        host.install_welcome_pane(welcome_message, on_service_welcome_login);
    }

    private void on_service_welcome_login() {
        debug("EVENT: user clicked 'Login' in welcome pane.");

        do_authentication();
    }
}

}

